<?php
$textpart = "relate";
include_once("tng_begin.php");
//v 11.0.0.1f

/**
*  Description of the bidirectional breadth-first algorithm:
* 
* Source and destination searches are recorded by the same array $visited (indexed by id) with source 
* starting at 1 and increasing and destination at -1 and decreasing. Each step of the search constructs 
* two arrays of the new people found on that step (one +ve, one -ve). There are two SQL queries: one 
* finds parents and siblings, the other spouses and children. Finding a non-zero entry of the
* opposite sign means the search is concluded whereas if it is of the same sign it may be discarded.
* Either of these can terminate the search although the step is completed in case another solution is 
* found.
* To handle genetic as well as inclusive searches, the search in each direction is split into two parts,
* those that search upward in the tree and those that search down. It is then possible to manipulate
* the lists of new people appropriately in both cases.
* If no new people are found on either of the searches, it fails immediately, so that an isolated 
* part of the tree doesn't lead to unecessary activity.
* For efficiency, the searches at every step are bundled together using an "in" clause in the SQL. 
* Constructing the solution starts with the duplicated node (stored in the $match variable) and fans out 
* towards each end using a back trace ($path) stored with each node. Relationships are also stored (in
* $relation) although they may need to e inverted.
* The relationship is then converted to a string of letters (p,i,s,c - parent, sibling, spouse, child) to 
* use pattern-matching to produce the final relationships.
*
* arguments: tree,from, setid, genetic, relatemax, forget
*/	

//These values can be set from outside
$limit=24;			#default distance (links) to search
$relclmax = 2;		#max no of relative clauses in a sentence.
$printrel=0;		#set to 1 to output coded relationship for checking
$timer=0;			#set to 1 to print time

global $visited, $famvisit, $queries, $times, $fail;
define('SPOUSE', 's');		#the basic relations
define('PARENT', 'p');		#spouse, parent, sibling, child
define('SIBLING', 'i');
define('CHILD', 'c');
define('BLOODREL',9);		#offset in text_settings array
define('ALWAYS',10);

#half of relationships are recorded in reverse so have to be inverted
$inv=array(SPOUSE=>SPOUSE, PARENT=>CHILD, SIBLING=>SIBLING, CHILD=>PARENT);	
$getperson_url = getURL( "getperson", 1 );
$start=cptime();
$queries=0; 
if (!$relatemax) $relatemax=1;

if (!($from || $setid) || !$tree) {
	echo "Error: Not enough data: from:$from; tree:$tree; setid:$setid;\n";
	exit();
}
$from = trim($from);
$setid = trim($setid);
$tree = trim($tree);

if ($genetic==1)							#remember settings of genetic
	$_SESSION['genetic'] = 1;
elseif ($genetic==2)						#2 turns off genetic
	$_SESSION['genetic'] = 0;
$genetic = $_SESSION['genetic'] ? 1: 0;		#use the new value

if ($setid) {
	if ($forget) {
		unset($_SESSION['anchorID']);
		unset($_SESSION['anchorTree']);
		$from = $setid;
	}
    else {
		$_SESSION['anchorID']= $setid;
		$_SESSION['anchorTree'] = trim($tree);
		dropdown($setid,$setid,$setid,0);
		echo "$text[anchorrelationships]<br/>\n";
    	exit();
    }
}

if(!$to) {      #the normal case. Find who to relate to
    if (caseq($_SESSION['anchorTree'],$tree) && $_SESSION['anchorID']) 
        $to = $anchor = trim($_SESSION['anchorID']);       #they've chosen an id in this tree
    elseif ($_SESSION['logged_in'] && caseq($_SESSION['mygedcom'],$tree) && $_SESSION['mypersonID'])
        $to = $anchor = trim($_SESSION['mypersonID']);     #use their own id
    elseif ($to = $anchor = $relateanchor[$tree]) ;	#has been set in customconfig by admin
    else {		#last resort choose first non-living person in the tree
		$number= $tngconfig['personprefix'] ? "abs(substring(personID,".(strlen($tngconfig['personprefix'])+1) ."))"
			: ($tngconfig['personsuffix'] ? "abs(substring(personID,1,length(personID)-".strlen($tngconfig['personsuffix'])."))"
			: "personID");						#selects the person number in the db
		$result = tng_query("select personID from $people_table where !living and gedcom='$tree' order by $number limit 1");
     	if($row =tng_fetch_array($result))           #there is a non-living person in this tree
        	$to = trim($row[0]);
    	else {
            echo $text['withthisperson'];
            $_SESSION['anchorID'] = $from;
            $_SESSION['anchorTree'] = $tree;
            exit();
        }
    }
}

#echo "<table width='100%'><tr><td>";
if (caseq($from,$to)) {
		dropdown($from,$to,$from,0);
		echo "$text[anchorrelationships]<br/>\n";
}
elseif ($found=relate($tree, $from,$to, $limit,$genetic)) {
	$res=0;
	foreach ($found as $pid) {				#reconstruct the path
		if ($test) echo "\n";
		list($pathid,$reln,$level)=$match[$pid];
		$list=array(); $distance=0;
		if ($level>0) { $a=$pathid; $b=$pid; }
		else { $a=$pid; $b=$pathid; $reln=$inv[$reln];}
		while(1) {
			array_unshift($list,getdetails($a,$tree,$relation[$a]));
			if (caseq($a,$from) || caseq($a,$path[$a])) break;
			$a=$path[$a];
			$distance++;
		}
		while (1) {
			array_push($list,getdetails($b,$tree,$reln));
			$reln=$inv[$relation[$b]];	#invert and remember for next loop	
			if (caseq($b,$to) || caseq($b,$path[$b])) break;
			$b=$path[$b];
			$distance++;
		}

		$strings[$res]='';
		for ($i=0; $i<count($list); $i++) {			#read off the short form
			list($relnm,$name1,$sex1, $id1) = $list[$i];
			$name[$res][$i] = $name1;
			if($i>0) {
				$reln= $rel[$relnm][$lastsex];	#need to use sex of previous person in chain
				if ($test) echo "$reln of ";
				$strings[$res] .= $relnm;
				$sex[$res][$i] = $lastsex;
				$birth[$res][$i]=$birth1;
			}
			if ($test) echo "$name1 \n";
			$lastsex=$sex1;
		}
		$res++;
	}
	
	$solutions = count($strings);
	dropdown($from,$to,$anchor,$solutions);

	for ($soln=0; $soln<min($relatemax,$solutions); $soln++) {		#construct text for each interpretation
		$string=$strings[$soln];
		if ($test) echo "$from $string<br/>\n";
		$p=1;
		$w=0;						#avoid too many "who is"s
		$output = "";
		if ($soln>0) echo "<p>"; 
		if ($printrel) echo $string.' ';
		$fname = getdetails($from,$tree,0);
		echo "$fname[4] "."$text[is] ";
		while ($string) {
			$loop=$string;
			if (preg_match('/^(c+)i(p+)/',$string,$m)) {
				$up=strlen($m[1]); $down=strlen($m[2]);
				if ($up>$down) {		#leave substantive part for next time
					$output .= relstring('child',$sex[$soln][$p],0,$up-$down-1);
					$string=substr($string,$up-$down);
					$p += $up-$down;
				}
				else {
					$output .= relstring('cousin', $sex[$soln][$p],0,$up-1);
					$string=substr($string,2*$up+1);
					$p += 2*$up+1;
				}
			}
			elseif (preg_match('/^(c+)i(s?)/',$string,$m)) {
				$up=strlen($m[1]);
				$across=strlen($m[2]);
				$output .= relstring('nibling',$sex[$soln][$p],0,$up-1);
				$bloodnephew = $text_setting['nibling'][BLOODREL];	#true in some languages
				if ($across && $bloodnephew) 
					$output .= relstring('spouse',$sex[$soln][$p+$up+1]);
				$string=substr($string,$up+$across+1);
				$p += $up+$across+1;
			}
			elseif (preg_match('/^i(p*)/',$string,$m)) {
				$down=strlen($m[1]);
				if ($down) $output .= relstring('uncleaunt',$sex[$soln][$p],0,$down-1);
				else $output .= relstring('sibling',$sex[$soln][$p]);			
				$string=substr($string,$down+1);
				$p+=$down+1;
			}
			elseif (preg_match('/^(c+)(p+)/',$string,$m)) {
				$up=strlen($m[1]); $down=strlen($m[2]);
				if ($up>$down) {
					$output .= relstring('child',$sex[$soln][$p],0,$up-$down-1);
					$string=substr($string,$up-$down);
					$p += $up-$down;
				}
				else {
					if ($up==1) $output .= relstring('sibling', $sex[$soln][$p],'half');
					else $output .= relstring('cousin', $sex[$soln][$p],'half',$up-2);
					$string=substr($string,$up*2);
					$p += 2*$up;
				}
			}
			elseif (preg_match('/^s(c+)i(p+)/',$string,$m)) {	#lookahead for cousin
				$output .= relstring('spouse',$sex[$soln][$p]);
				$string=substr($string,1);
				$p += 1;				
			}
			elseif (preg_match('/^(sc|ps)/',$string,$m)) {
				$output .= relstring($m[1][0]=='s' ? 'child':'parent',$sex[$soln][$p],'inlaw');			
				$string=substr($string,2);
				$p+=2;
			}
			elseif (preg_match('/^si(p+)/',$string,$m)) {
				$down = strlen($m[1]);
				$blooduncle = $text_setting['uncleaunt'][BLOODREL];	#true in some languages
				if ($blooduncle) {
					$output .= relstring('spouse',$sex[$soln][$p]);
					$p++;
				}
				$output .= relstring('uncleaunt',$sex[$soln][$p],0,$down-1);			
				$string=substr($string,$down+2);
				$p+=$down+($blooduncle?1:2);
			}
			elseif (preg_match('/^(p+)/',$string,$m)) {
				$dist=strlen($m[1]);
				$output .= relstring('parent',$sex[$soln][$p],0,$dist-1);
				$string=substr($string,$dist);
				$p+=$dist;
			}
			elseif (preg_match('/^(c+)s(p+)/',$string,$m)) {
				$up=strlen($m[1]);
				$down=strlen($m[2]);
				if ($down==1 && $up>1) {
					$output .= relstring('child',$sex[$soln][$p],0,$up-2);
					$string=substr($string,$up-1);
					$p += $up-1;
				}
				elseif ($up==1 && $down==1) {
					$output .= relstring('sibling', $sex[$soln][$p],'step');
					$string=substr($string,3);
					$p += 3;
				}
				else {		#$up==1 && $down>1
					$output .= relstring('child',$sex[$soln][$p],0,0);
					$string=substr($string,1);
					$p += 1;
				}
			}
			elseif (preg_match('/^((c+)s|s(p+))/',$string,$m)) {
				$up=strlen($m[2]);
				$down=strlen($m[3]);
				if ($up==1 || $down>0) {
					$output .= relstring(($up==1 ? 'child':'parent'),$sex[$soln][$p],'step');
					$string=substr($string,2);
					$p +=2;
				}
				elseif ($up>1) {
					$output .= relstring('child',$sex[$soln][$p],0,$up-2);
					$string=substr($string,$up-1);
					$p += $up-1;
				}
			}
			elseif (preg_match('/^(c+)/',$string,$m)) {
				$dist=strlen($m[1]);
				$output .= relstring('child',$sex[$soln][$p],0,$dist-1);
				$string=substr($string,$dist);
				$p+=$dist;
			}
			elseif (preg_match('/^(sis?|isi?)/',$string,$m)) {
				$dist=strlen($m[1]);
				$output .= relstring('sibling',$sex[$soln][$p],'inlaw');			
				$string=substr($string,$dist);
				$p+=$dist;
			}
			elseif (preg_match('/^s/',$string,$m)) {
				$output .= relstring('spouse',$sex[$soln][$p]);			
				$string=substr($string,1);
				$p+=1;
			}
			$w++;
			$output .= $name[$soln][$p-1];
			$pronoun = caseq($sex[$soln][$p],'F')?$text['she']:$text['he']; 
			if ($string) $output .= $w>$relclmax ? ".<br/>\n$pronoun $text[is] " : 
				/*($w==1 ? ", ": */ ", $text[whois] ";
			if ($w>$relclmax) $w=0;
			if ($string==$loop) break;
		}
		echo $output;
		echo $soln>0 ?".</p>\n":".<br/>\n";
		if ($string) echo "*** missed relations: $string ***<br/>\n";
	}
}
else {		#search failed
	dropdown($from,$to,$anchor,0);	
	$fname = getdetails($from,$tree,0);
	$tname = getdetails($to,$tree,0);
	echo "$fname[4] "."$text[is] ";
	if (!$fname[3]) echo " $text[notvalid].<br/>\n";
	elseif (!$tname[3]) echo "$to $text[notvalid].<br/>\n";
	elseif (list($id,$lev)=$fail) {
		if ($genetic) echo "$fname[1] $text[notgeneticallyrelatedto] $tname[1]<br/>\n";
		else echo (caseq($id,$from)?$fname[1]:$tname[1])." $text[notconnectedto].<br/>\n";
	}
	else echo str_replace('xxx',$limit,$text['notrelatedto']).$tname[1]."<br/>\n";
}
$time = round((cptime()-$start)/1000,3);
if ($timer) echo " $time secs";

/**
*	dropdown generates the options for the next move.
*	It takes advantage of the tngdd css class so that styling corresponds with template
*	(but doesn't use the tinydropdown javascript routine)
**/
function dropdown($from,$to,$anchor,$solutions) {	#
	global $text,$relateanchor,$genetic,$relatemax,$tree;
	echo "<div style='float:right;'><ul class='tngdd'><li class='rellink'><a href='#' class='menulink'>$text[more]</a>\n";
	echo "<ul class='relmenu'>\n";	
	if ($solutions>$relatemax) 
		echo ddrow("togglelink('$from','$tree',$solutions)",'showmorelines');
	if ($from!=$to )
		echo ddrow("makeroot('$from','$tree',0)",'anchormake');
	if ($_SESSION['anchorID'] && caseq($_SESSION['anchorTree'],$tree))
		echo ddrow("makeroot('$from','$tree', 1)",'anchorreset');
	if ($from != $anchor) {
		if ($genetic) echo ddrow("genetic('$from','$tree',2)",'includenongenetic');
		else echo ddrow("genetic('$from','$tree',1)",'showgenetic');
	}
	if ($_SESSION['tngrole'] == "admin") {
		if ($from != $relateanchor[$tree])
			echo ddrow("makeglobal('$from','$tree',0)",'makedefault');
		else echo ddrow("makeglobal('$from','$tree',1)",'forgetdefault');
	}
	echo "</ul></li></ul></div>\n";
}

function ddrow($action,$key) {
global $text;
return "<li><span class='menu-label'><a href='#' onclick=\"$action\">$text[$key]</a></li></span>\n";
}

/**
*	getdetails returns the name and sex for personID in array, together with relation & id
*/
function getdetails($personID, $tree, $reln) {
	global $people_table, $getperson_url;
	$result = tng_query("select * from $people_table where personID='$personID' and gedcom='$tree'") ;
	$row=tng_fetch_assoc($result);
	$righttree = checktree($tree);
	$rightbranch = $righttree ? checkbranch($row['branch']) : 0;
	$rights=determineLivingPrivateRights($row, $righttree, $rightbranch);
	$row['allow_living'] = $rights['living'];
	$row['allow_private'] = $rights['private'];
	$row['both'] = $rights['both'];
	$displayname = getName($row);    #respects privacy
	$name = "<a href = '{$getperson_url}personID=$personID&tree=$tree'>$displayname</a>";
	return array($reln,$name, $row['sex'],$personID,$displayname);	#relation, name, sex, id
}

/**
 * relstring constructs a string which expresses a relation
 * e.g. relstring('cousin','M','half',2,0) -> the 2 x half cousin of
 * @returns string
*/
function relstring($generic,$sex,$discr=0,$gen=0) {
	global $text, $text_setting;
	#basic relationships by gender
	$tng_strings = array( 
		'basic'=> array(
			'child'=>array('M'=>'son','F'=>'daughter','U'=>'rchild'),
			'parent'=>array('M'=>'fathof','F'=>'mothof','U'=>'parof'),
			'cousin'=>array('M'=>'mcousin','F'=>'fcousin','U'=>'cousin'),
			'sibling'=>array('M'=>'brother','F'=>'sister','U'=>'sibling'),
			'spouse'=>array('M'=>'rhusband','F'=>'rwife','U'=>'rspouse'),
			'nibling'=>array('M'=>'nephew', 'F'=>'niece','U'=>'nephnc'),
			'uncleaunt'=>array('M'=>'uncle','F'=>'aunt','U'=>'uncleaunt')),
		'inlaw'=> array(
			'child'=>array('M'=>'sil','F'=>'dil','U'=>'sdil'),
			'sibling'=>array('M'=>'bil', 'F'=>'ssil', 'U'=>'bsil'),
			'parent'=>array('M'=>'fil','F'=>'mil','U'=>'fmil')),
		'half' => array(
			'sibling'=>array('M'=>'halfbrother','F'=>'halfsister','U'=>'halfsibling'),
			'cousin'=>array('M'=>'mhalfcousin','F'=>'fhalfcousin','U'=>'halfcousin')),
		'step' =>array(
			'child'=>array('M'=>'stepson','F'=>'stepdau','U'=>'stepchild'),
			'parent'=>array('M'=>'stepfath','F'=>'stepmoth','U'=>'steppar'),
			'sibling'=>array('M'=>'stepbrother','F'=>'stepsister','U'=>'stepsibling')));
	$gen1 = array( 
		array('child'=>array('M'=>'gson','F'=>'gdau','U'=>'gsondau'),
			'parent'=>array('M'=>'gfath','F'=>'gmoth','U'=>'gpar'),
			'nibling'=>array('M'=>'gnephew','F'=>'gniece','U'=>'gnephnc'),
			'uncleaunt'=>array('M'=>'guncle','F'=>'gaunt','U'=>'guncleaunt')),
		array('child'=>array('M'=>'ggson','F'=>'ggdau','U'=>'ggsondau'),
			'parent'=>array('M'=>'ggfath','F'=>'ggmoth','U'=>'ggpar')));
	$tng_defs = array('parent'=>2, 'child'=>2, 'cousin'=>0,'nibling'=>1, 'uncleaunt'=>1);
	if (!$sex) $sex = 'U'; 		//if not in db
	if (!$discr) $discr = 'basic';
	$emax=$tng_defs[$generic];
	if ($gen==0||$generic=='cousin') $key = $tng_strings[$discr][$generic][$sex];	# -> son, stepson, cousin etc
	else $key=$gen1[min($gen,$emax)-1][$generic][$sex];	# -> gson, ggson etc.
	$modifier= $gen>$emax ? ($gen-$emax+1).' x ':'';		#the default
	$prefix = 'xxx ';
	if (in_array($generic, array('parent','child','cousin','nibling','uncleaunt')) 
		&& $text_setting[$generic]) {	#language settings in text file
		list($terms,$maxord,$diff,$post,$ord,$m1,$f1,$m2,$f2,$bl,$all) = $text_setting[$generic];
		if ($terms) {					#establish a different key
			if ($terms<$emax+1 && !$all) $key= $terms==1?$tng_strings[$discr][$generic][$sex] 
				:$gen1[0][$generic][$sex];	#Portuguese avô/avó
			elseif (($gen>$emax) || $all) $key .= min($gen+1,$terms);			#produces ggfath3 mcousin2 etc
		}
		else $terms=$emax+1;
		if ($gen+$diff>=$terms) { 		#need an extra ordinal or multiplier
			$n = $gen-$terms+2+$diff+$diff2;
			if ($n<=$maxord) $modifier = $text[$ord.$n].(caseq($sex,'F')?$f2:$m2); #second, segunda etc
			else {
				$term = caseq($sex,'F')?$f1:$m1;
				if (strpos($term,'xxx ')) $replace=str_replace('xxx ',$n,$term);
				else $replace = $n.$term;		#6th or 10&ordf; 
			}
		}
		else $modifier='';
		if ($post) $prefix='yyy';
	}
	$result = $text[$key];
	if ($gen) $result = str_replace($prefix, $modifier, $result);
	$result=str_replace(array('xxx ','yyy '),array('',''),$result);			#clean up
	if (!$result)
		echo "**** no text for $generic, $sex, $discr, $gen, $key ****\n";
	return $result.' ';
}

/**
*	relate performs the search. If $target is null, then it simply finds all relatives to
*	the defined limit. Otherwise it stops when a match is found.
*   @returns mixeds[]; On failure, $fail is set
**/
function relate( $tree, $start, $target, $limit, $genetic) {
	global $visited, $famvisit, $path, $relation, $match, $fail;
	$times[1]=cptime();
	$visited = array($start=> 1, $target=> -1);
	$forup=$fordown= array($start); 
	$backup=$backdown= array($target);
	$famvisit = $path = $relation = $match = array();
	$level =1; $found=array();
	while (!$found && $level<$limit) {
		$level++;
		$newforup=getup($forup,$tree, $level,$genetic);
		$newfordown=getdown($fordown,$tree,  $level,$genetic);
		$newfor= array_merge($newforup,$newfordown);
		if (!$newfor) {$fail=array($start,$level); return 0;}	#failure
		if ($target) {
			if(!is_array($backdown)) print "backdown=$backdown;";
			$found=array_intersect($newfor, array_merge($backup,$backdown));
			if ($found) break;
			$newbackup=getup($backup,$tree,  -$level,$genetic);
			if (!$genetic) {		
				$newbackdown=getdown($backdown,$tree,  -$level,$genetic);
				$newback = array_merge($newbackup,$newbackdown);
				if (!$newback) {$fail=array($target,$level); return 0;}		#failure		
				$found=array_intersect($newback, array_merge($forup,$fordown,$newfor));
			}
			else { $found = array_keys($match);		
				if(!$newforup && !$newbackup) {$fail=array($target,$level); return 0;}
				$newback=$newbackup;
			}
		}
		$forup = $genetic?$newforup:$newfor;
		$fordown=$newfor;
		$backup= $genetic?$newbackup:$newback;
		$backdown=$newback;
	}
	return $found;
}

/**
 *	getup finds all parents and siblings of the current set of persons that haven't been previously
 *	visited. If $genetic is set, it doesn't include parents whose relationship is not birth (or blank)
**/
function getup($persons, $tree, $level,$genetic) {	#visit parental families
	global $visited, $famvisit, $path, $relation, $match, $nonbirth, $families_table, $children_table;
	$ids = implode("','",$persons);
	$result=array();
	$res = tng_query("SELECT a.husband, a.wife, c.personID, b.personID, a.familyID,b.frel,b.mrel
		FROM $families_table a join $children_table b on a.familyID=b.familyID and a.gedcom=b.gedcom 
		join $children_table c on b.familyID=c.familyID and b.gedcom=c.gedcom
		WHERE a.gedcom='$tree' and b.personID in ('$ids')");
	while ($row=tng_fetch_array($res)) {
		$personID = $row[3];
		for ($i=0;$i<=2; $i++) {
			$r=$row[$i];
			$rln=caseq($row[$i+5],'birth')?'':$row[$i+5];
			if ($i==0) $famvisit[$r]=$row[4];
			if ($r && ($level<0 ? $visited[$r]>=0 : $visited[$r]<=0) && ($genetic? !$rln :true)) {
				if ($visited[$r] ) {		#been visited in other direction, save path
					  if(!$match[$r]) {
						  $match[$r]=array($personID,($i==2 ? SIBLING:CHILD),$level,$rln);
						  array_push($result,$r);
					  }
				}
				else {
					$visited[$r] = $level;				#record visit
					$path[$r] = $personID;				#backtrace
					$relation[$r] = ($i==2 ? SIBLING:CHILD);
					array_push($result,$r);
				}
			}
		}
	}
	return $result;
}

/**
 *	getdown finds all spouses and children of the current set of persons that haven't been previously
 *	visited. If $genetic is set, it doesn't include spouses or children whose relationship is 
 *	not birth (or blank)
**/
function getdown($persons, $tree, $level, $genetic ) {	#visit descendant families
	global $visited, $famvisit, $path, $relation, $match, $nonbirth, $families_table, $children_table;
	$ids = implode("','",$persons);
	$result=array();
	$res = tng_query("SELECT a.husband, a.wife, b.personID, a.husband in ('$ids'), a.familyID,b.frel,b.mrel
		FROM $families_table a LEFT JOIN $children_table b ON a.familyID = b.familyID AND a.gedcom = b.gedcom
		WHERE a.gedcom='$tree' and (a.husband in ('$ids') or a.wife in ('$ids'))");
	while ($row=tng_fetch_array($res)) {
		$personID = $row[3] ? $row[0]:$row[1];
		for ($i=0;$i<=2; $i++) {
			$r=$row[$i];
			$rln=strcasecmp($row[$i+5],'birth')?'':$row[$i+5];
			if (!$genetic||$i==2) {			#omit spouse if genetic
				if ($r && ($level<0 ? $visited[$r]>=0 : $visited[$r]<=0) && ($genetic? !$rln:true)) {
					if ($visited[$r] ) {		#found a match, save path
						  if(!$match[$r]) {
							  $match[$r]=array($personID,($i==2 ? PARENT:SPOUSE),$level,$rln);
							  array_push($result,$r);
						  }
					}
					else {
						$visited[$r] = $level;
						$path[$r] = $personID;				#backtrace
						$relation[$r] = $i==2 ? PARENT:SPOUSE;	
						$nonbirth[$r]=$rln;
						array_push($result,$r);
					}
				}
			}
		}
	}
	return $result;
}

/**
*	caseq trims both arguments and does a case insensitive equality test
*/
function caseq($a,$b) {
	return strcasecmp(trim($a),trim($b))==0;
}

function cptime() {	#ms fp
	return gettimeofday(true)*1000;
}
